#include <HALOptions.h>
#include <SerialChipRegistry.h>

#include "CircleBuf.h"
#include "BluntServer.h"
#include "HCI.h"
#include "L2CAP.h"
#include "RFCOMM.h"
#include "SDP.h"

#define kReceiveBufReg 3
#define kTransmitBufReg 3
#define kInterruptEnbReg 2
#define kInterruptIDReg 1
#define kLineControlReg 0
#define kModemControlReg 7
#define kLineStatusReg 6
#define kModemStatusReg 5
#define kScratchReg 4

#define DEBUGSTR
// #define DEBUGSTR DebugStr
// #define DEBUGSTR printf

#define DMA 0

extern "C" void EnterFIQAtomic (void);
extern "C" void ExitFIQAtomic (void);
extern "C" void _EnterFIQAtomic (void);
extern "C" void _ExitFIQAtomicFast (void);
extern TUPort* GetNewtTaskPort (void);
extern "C" long LockStack(TULockStack* lockRef, ULong additionalSpace);	// Lock the entire stack plus additionalSpace # of bytes
extern "C" long UnlockStack(TULockStack* lockRef);

// ================================================================================
//  Interrupt Handlers
// ================================================================================

void BluntServer::RxCSpecialIntHandler (void)
{
}

void TxBEmptyIntHandler (void* server)
{
	DEBUGSTR ("TxBEmptyIntHandler\n");
	((BluntServer*)server)->TxBEmptyIntHandler ();
}

void ExtStsIntHandler (void* server)
{
	DEBUGSTR ("ExtStsIntHandler\n");
//	 ((BluntServer*)server)->ExtStsIntHandler ();
}

void RxCAvailIntHandler (void* server)
{
	DEBUGSTR ("RxCAvailIntHandler\n");
	 ((BluntServer*)server)->RxCAvailIntHandler ();
}

void RxCSpecialIntHandler (void* server)
{
	DEBUGSTR ("RxCSpecialIntHandler\n");
//	 ((BluntServer*)server)->RxCSpecialIntHandler ();
}

void RxDmaIntHandler (void* server, RxErrorStatus status)
{
	DEBUGSTR ("RxDmaIntHandler\n");
	 ((BluntServer*)server)->RxDMAIntHandler (status);
}

void TxDmaIntHandler (void* server)
{
	DEBUGSTR ("TxDmaIntHandler\n");
	 ((BluntServer*)server)->TxDMAIntHandler ();
}

#pragma mark -

// ================================================================================
//  Driver Specific Functions
// ================================================================================

void BluntServer::DriverReset (void)
{
	TSerialChip16450 *chip;
	Byte prescale;
	
	LOG ("BluntServer::DriverReset (%d)\n", fDriver);
	chip = (TSerialChip16450 *) fChip;
	if (fDriver == DRIVER_BT2000) {
		// Enable Enhanced Mode to get larger FIFO buffers
		chip->WriteSerReg (kLineControlReg, 0x83);
		chip->WriteSerReg (kInterruptIDReg, 0x21);
		chip->WriteSerReg (kLineControlReg, 0x03);
	} else if (fDriver == DRIVER_BT2000E ||
		fDriver == DRIVER_950_7MHZ ||
		fDriver == DRIVER_950_14MHZ ||
		fDriver == DRIVER_950_18MHZ ||
		fDriver == DRIVER_950_32MHZ) {
		// Enable '950 Mode
		chip->WriteSerReg (kLineControlReg, 0xbf);
		chip->WriteSerReg (kInterruptIDReg, 0x10);
		chip->WriteSerReg (kLineControlReg, 0x03);

		// Enable Enhanced Mode to get larger FIFO buffers
		chip->WriteSerReg (kLineControlReg, 0x83);
		chip->WriteSerReg (kInterruptIDReg, 0x21);
		chip->WriteSerReg (kLineControlReg, 0x03);

		// Enable the prescaler
		chip->WriteSerReg (kModemControlReg, 0x80);
		chip->WriteSerReg (kScratchReg, 0x01);
		chip->WriteSerReg (kLineStatusReg, 0x20);

		// Set the prescale value
		if (fDriver == DRIVER_BT2000E || fDriver == DRIVER_950_7MHZ) {
			prescale = 0x20;
		} else if (fDriver == DRIVER_950_14MHZ) {
			prescale = 0x40;
		} else if (fDriver == DRIVER_950_18MHZ) {
			prescale = 0x50;
		} else if (fDriver == DRIVER_950_32MHZ) {
			prescale = 0x8b;
		}
		chip->WriteSerReg (kScratchReg, 0x01);
		chip->WriteSerReg (kLineStatusReg, prescale);
	}
	Wait (100);
}

void BluntServer::DriverSendDelay (void)
{
	switch (fDriver) {
		case DRIVER_TAIYO_YUDEN:
			Wait (1);
			break;
		default:
			break;
	}
}

#pragma mark -

// ================================================================================
//  Generic Protocol Handler Specific Functions
// ================================================================================

Handler::Handler ()
{
	fNumHandlers = 0;
	fServer = NULL;
	fParentHandler = NULL;
}

Handler::~Handler ()
{
}

void Handler::AddHandler (Handler *handler, BluntServer *server)
{
	if (fNumHandlers < MAX_HANDLERS) {
		fHandlers[fNumHandlers] = handler;
		fHandlers[fNumHandlers]->fServer = server;
		handler->fParentHandler = this;
		fNumHandlers++;
	} else {
		printf ("**** %08x number of handlers exceeded ****\n", this);
	}
}

void Handler::RemoveHandler (Handler *handler)
{
	int i;
	
	for (i = 0; i < fNumHandlers; i++) {
		if (fHandlers[i] == handler) {
			memmove (&fHandlers[i], &fHandlers[i + 1],
				sizeof (Handler *) + fNumHandlers - 1);
			fNumHandlers--;
			break;
		}
	}
}

void Handler::Print (int depth)
{
	int i, j;
	char buffer[80];
	
	for (j = 0; j < depth; j++) buffer[j] = ' ';
	buffer[j] = 0;
	if (fHandlerType == H_HCI) LOGX ("%sHCI (%04x)", buffer, ((HCI *) this)->fConnectionHandle);
	else if (fHandlerType == H_L2CAP) LOGX ("%sL2CAP", buffer);
	else if (fHandlerType == H_SDP) LOGX ("%sSDP", buffer);
	else if (fHandlerType == H_RFCOMM) LOGX ("%sRFCOMM", buffer);
	LOGX (" (%08x)\n", this);
	for (i = 0; i < fNumHandlers; i++) {
		fHandlers[i]->Print (depth + 4);
	}
}

#pragma mark -

// ================================================================================
//  Server Functions
// ================================================================================

NewtonErr BluntServer::Initialize (ULong location, ULong driver)
{
	PSerialChipRegistry* reg;
	SerialChipID id;
	NewtonErr r = noErr;

	fLogLevel = 1;
	LOG ("BluntServer::Initialize\n");
	fLocation = location;
	fDriver = driver;
	fChip = NULL;
	reg = GetSerialChipRegistry ();
	id = reg->FindByLocation (fLocation);
	fChip = reg->GetChipPtr (id);
	LOG ("Driver: %d, Chip: %d (%08x)\n", fDriver, id, fChip);
	if (fChip == NULL) {
		LOG ("No chip found at this location!\n");
		r = -100000;
	} else {
		* (Long *) ((Byte*) (fChip) + 0x10) = 0;
	}
	return r;
}

ULong BluntServer::GetSizeOf ()
{
	return sizeof (BluntServer);
}

void BluntServer::TxBEmptyIntHandler (void)
{
	ULong n, size;
	int i;
	UByte c;
		
	if (fTxDMABuffer->BufferCount () > 0) {
		EnterFIQAtomic ();
		if (DMA) {
			fChip->TxDMAControl (kDMAStart);
		} else {
			fChip->SetTxDTransceiverEnable (true);
			while (fChip->TxBufEmpty () && fTxDMABuffer->BufferCount () > 0) {
				fTxDMABuffer->GetNextByte (&c);
				fChip->PutByte (c);
			}
		}
		fOutputCurrent += size - n;
		ExitFIQAtomic ();
	} else {
		EnterFIQAtomic ();
		fChip->ResetTxBEmpty ();
		ExitFIQAtomic ();
	}
}

void BluntServer::ExtStsIntHandler (void)
{
}

void BluntServer::RxCAvailIntHandler (void)
{
	ULong s;
	UByte c;
	ULong n;
	
	while (fChip->RxBufFull () && fInputCurrent < MAX_INPUT) {
		s = fChip->GetRxErrorStatus ();
		if (s != 0) fRxStatus = s;
		c = fChip->GetByte ();
		n = 1;
		fRxDMABuffer->CopyIn (&c, &n, false, 0);
	}
	SendForInterrupt (fPort, fIntMessage.GetMsgId (), 0, NULL, 0, M_DATA);
}

void BluntServer::RxDMAIntHandler (RxErrorStatus status)
{
	fRxErrorStatus = status;
	fChip->RxDMAControl (kDMANotifyOnNext | kDMASync);
	SendForInterrupt (fPort, fIntMessage.GetMsgId (), 0, NULL, 0, M_DATA);
}

void BluntServer::TxDMAIntHandler ()
{
}

long BluntServer::TaskConstructor ()
{
	TUNameServer nameServer;
	TCMOSerialIOParms parms;
	SCCChannelInts handlers;
	NewtonErr r;
	ULong n;
	HCI* hci;
	TULockStack lock;
	
	LOG ("BluntServer::TaskConstructor\n");
	fInputCurrent = 0;
	memset (fInputBuffer, 0, 128);
	fOutputCurrent = 0;
	fOutputCount = 0;

	fRxDMABuffer = new TExtendedCircleBuf ();
	fRxDMABuffer->Allocate (512, 8, 2, 0);
	fRxDMABuffer->DMABufInfo (&n, NULL, NULL, NULL);
	fTxDMABuffer = new TExtendedCircleBuf ();
	fTxDMABuffer->Allocate (512, 8, 2, 0);
	fTxDMABuffer->DMABufInfo (&n, NULL, NULL, NULL);
	fChip->SetSerialMode (kSerModeAsync);

	LOG ("Server: %08x Input: %08x Buffer: %08x\n", this, fInputBuffer, fRxDMABuffer);

	if (!fChip->PowerIsOn ()) fChip->PowerOn ();
	LOG ("Power: %d\n", fChip->PowerIsOn ());
	parms.fStopBits = k1StopBits;
	parms.fParity = kNoParity;
	parms.fDataBits = k8DataBits;
	parms.fSpeed = k115200bps;
	fChip->SetIOParms (&parms);

	fChip->SetInterruptEnable (false);
	LockStack (&lock, 64);
	EnterFIQAtomic ();
	fChip->Reconfigure ();
	if (DMA) {
		fChip->InitRxDMA (fRxDMABuffer, 129, RxDmaIntHandler);
		fChip->InitTxDMA (fTxDMABuffer, TxDmaIntHandler);
		fChip->RxDMAControl (kDMANotifyOnNext | kDMAStart);
	}
	fChip->SetTxDTransceiverEnable (true);
	ExitFIQAtomic ();
	UnlockStack (&lock);
	DriverReset ();

	LOG ("Chip configured\n");
	
	handlers.TxBEmptyIntHandler = ::TxBEmptyIntHandler;
	handlers.ExtStsIntHandler = ::ExtStsIntHandler;
	handlers.RxCAvailIntHandler = ::RxCAvailIntHandler;
	handlers.RxCSpecialIntHandler = ::RxCSpecialIntHandler;
	
	LOG ("TxBEmptyIntHandler: %08x\n",  &::TxBEmptyIntHandler);
	LOG ("ExtStsIntHandler: %08x\n",  &::ExtStsIntHandler);
	LOG ("RxCAvailIntHandler: %08x\n",  &::RxCAvailIntHandler);
	LOG ("RxCSpecialIntHandler: %08x\n",  &::RxCSpecialIntHandler);

	r = fChip->InstallChipHandler (this, &handlers);
	LOG ("InstallHandlers: %d\n", r);

//	Debugger();
//	LOG ("Server code: %08x\n", *((int *) &BluntServer::TxBEmptyIntHandler));
	fChip->SetIntSourceEnable (kSerIntSrcTxBufEmpty, true);
	fChip->SetInterruptEnable (true);
	
	fNumHandlers = 0;
	fHandlers = NULL;
	hci = new HCI ();
	hci->fConnectionHandle = -1;
	AddHandler ((Handler *) hci);
	
	fPort.Init ();
	fIntMessage.Init (false);
	LOG ("Server port: %08x\n", fPort.fId);
	nameServer.RegisterName ("BluntServer", "TUPort", fPort.fId, 0);
	
	fOutstandingPackets = 0;
	
	return noErr;
}

void BluntServer::TaskDestructor ()
{
	TUNameServer nameServer;
	int i;
	
	LOG ("BluntServer::TaskDestructor\n  Got %d bytes\n  ", fInputCurrent);
	for (i = 0; i < fInputCurrent && i < 256; i++) {
		LOG ("%02x ", fInputBuffer[i]);
		if ((i + 1) % 16 == 0) LOG ("\n  ");
	}
	LOG ("\n");
	* (Long *) ((Byte*) (fChip) + 0x10) = 0;
	fChip->PowerOff ();
	nameServer.UnRegisterName ("BluntServer", "TUPort");
}

void BluntServer::TaskMain ()
{
	ULong n = 0;
	ULong type = 0;
	Boolean end = false;
	TUMsgToken token;
	
	LOG ("BluntServer::TaskMain\n");
	while (!end) {
		n = *(ULong *) &::TxBEmptyIntHandler;
		n += *(ULong *) &::ExtStsIntHandler;
		n += *(ULong *) &::RxCAvailIntHandler;
		n += *(ULong *) &::RxCSpecialIntHandler;

		n = *(ULong *) &BluntServer::TxBEmptyIntHandler;
		n += *(ULong *) &BluntServer::ExtStsIntHandler;
		n += *(ULong *) &BluntServer::RxCAvailIntHandler;
		n += *(ULong *) &BluntServer::RxCSpecialIntHandler;
		n = 0;

		fPort.Receive (&n, fMessage, MAX_MESSAGE, &token, &type);
		LOG2 ("Got the message %d (%08x): %d bytes (%08x) \n", type, fMessage, n, &token);
		switch (type) {
			case M_DATA:
				n = sizeof (fInputBuffer) - fInputCurrent;
				fRxDMABuffer->CopyOut (fInputBuffer + fInputCurrent, &n, &fRxStatus);
				fChip->RxDMAControl (kDMANotifyOnNext | kDMAStart);
				fLastRead = sizeof (fInputBuffer) - fInputCurrent - n;
				fInputCurrent += fLastRead;
				HandleData ();
				break;
			case M_COMMAND:
				HandleCommand ((BluntCommand *) fMessage);
				break;
			case M_TIMER:
				HandleTimer ((BluntTimerEvent*) fMessage);
				break;
			case M_QUIT:
				end = true;
				break;
			default:
				break;
		}
	}
}

void BluntServer::HandleData ()
{
	int i;
	ULong lastAmount;
	
	LOG2 ("BluntServer::HandleData %08x %d\n", fInputBuffer, fInputCurrent);
	lastAmount = 0;
	do {
		lastAmount = fInputCurrent;
		for (i = 0; fInputCurrent > 0 && i < fNumHandlers; i++) {
			LOG2 ("BluntServer Handler: %d (%08x), data: %d bytes\n", i, fHandlers[i], fInputCurrent);
			if (fHandlers[i]->fConnectionState != CONN_REMOVE) {
				((HCI *) fHandlers[i])->HandleData (fInputBuffer, fInputCurrent);
			}
		}
		if (fInputCurrent > 0  && lastAmount == fInputCurrent) {
			((HCI *) fHandlers[0])->RemoveUnhandledPackets (fInputBuffer, fInputCurrent);
		}
	} while (fInputCurrent > 0 && lastAmount != fInputCurrent);
	
	i = 0;
	while (i < fNumHandlers) {
		if (fHandlers[i]->fConnectionState == CONN_REMOVE) {
			RemoveHandler (fHandlers[i]);
		} else {
			i++;
		}
	}
}
void BluntServer::ConsumeData (ULong amount)
{
	LOG2 ("BluntServer::ConsumeData %d (%d)\n", amount, fInputCurrent);

	if (fInputCurrent >= amount) {
		memcpy (fInputBuffer, fInputBuffer + amount, fInputCurrent - amount);
		fInputCurrent -= amount;
	} else {
		fInputCurrent = 0;
	}
}

void BluntServer::HandleCommand (BluntCommand* command)
{
	HCI* hci;
	int i;

	LOG ("BluntServer::HandleCommand %d\n", command->fType);
	
	switch (command->fType) {
		case C_RESET: {
			BluntResetCommand* c = (BluntResetCommand*) command;
			LOG ("  Reset: %s\n", c->fName);
			strcpy (((HCI *) fHandlers[0])->fName, c->fName);
			((HCI *) fHandlers[0])->Reset ();
			} break;
		case C_CONNECT: {
			BluntConnectionCommand* c = (BluntConnectionCommand*) command;
			LOG ("  Connect: %d\n", c->fTargetLayer);
			SetupProtocolStack (c);
			} break;
		case C_DISCONNECT: {
			BluntDisconnectCommand* c = (BluntDisconnectCommand*) command;
			LOG ("  Disconnect: %04x\n", c->fHCIHandle);
			for (i = 1; i < fNumHandlers; i++) {
				hci = (HCI *) fHandlers[i];
				if (hci->fHandlerType == H_HCI /* XXX &&
					hci->fConnectionHandle == ((BluntDisconnectCommand *) command)->fHCIHandle */) {
					hci->Disconnect (0x13);
					DeferredRemoveHandler (hci);
				}
			}
			} break;
		case C_INQUIRY: {
			BluntInquiryCommand* c = (BluntInquiryCommand*) command;
			LOG ("  Inquiry %d %d\n", c->fTime, c->fAmount);
			((HCI *) fHandlers[0])->Inquiry (c->fTime, c->fAmount);
			} break;
		case C_INQUIRY_CANCEL: {
			LOG ("  Cancel inquiry\n");
			((HCI *) fHandlers[0])->InquiryCancel ();
			} break;
		case C_NAME_REQUEST: {
			BluntNameRequestCommand* c = (BluntNameRequestCommand*) command;
			LOG ("  Name request\n");
			((HCI *) fHandlers[0])->RemoteNameRequest (c->fBdAddr, c->fPSRepMode, c->fPSMode);
			} break;
		case C_INIT_PAIR:
			LOG ("  Pair\n");
			InitiatePairing ((BluntInitiatePairingCommand*) command);
			break;
		case C_SERVICE_REQUEST:
			LOG ("  Service request\n");
			InitiateServiceRequest ((BluntServiceRequestCommand*) command);
			break;
	}
//	delete command->fOriginalCommand;
}

void BluntServer::HandleTimer (BluntTimerEvent *event)
{
	Handler *handler = event->fHandler;
	LOG ("BluntServer::HandleTimer %08x\n", handler);
	switch (handler->fHandlerType) {
		case H_HCI:
			((HCI *) handler)->HandleTimer (event->fUserData);
			break;
		case H_L2CAP:
			((L2CAP *) handler)->HandleTimer (event->fUserData);
			break;
		case H_SDP:
			((SDP *) handler)->HandleTimer (event->fUserData);
			break;
		case H_RFCOMM:
			((RFCOMM *) handler)->HandleTimer (event->fUserData);
			break;
	}
	delete event->fOriginalEvent;
}

void BluntServer::SetupProtocolStack (BluntConnectionCommand* command)
{
	HCI* hci;
	L2CAP* l2cap;
	RFCOMM* rfcomm;
	SDP* sdp;
	
	LOG ("BluntServer::SetupProtocolStack\n");
	fNumHandlers = 1;

	hci = SetupHCILayer (command->fToolPort, command->fBdAddr, command->fPSRepMode, command->fPSMode);
	switch (command->fTargetLayer) {
		case L_HCI:
			hci->Connect ();
			break;
		case L_L2CAP:
			l2cap = SetupL2CAPLayer (command->fToolPort, command->fL2CAPProtocol, hci);
			l2cap->Connect ();
			break;
		case L_RFCOMM:
			l2cap = SetupL2CAPLayer (command->fToolPort, PSM_RFCOMM, hci);
			rfcomm = SetupRFCOMMLayer (command->fToolPort, command->fRFCOMMPort, l2cap);
			rfcomm->Connect ();
			break;
		case L_SDP:
			l2cap = SetupL2CAPLayer (command->fToolPort, PSM_SDP, hci);
			sdp = SetupSDPLayer (command->fToolPort, l2cap);
			sdp->Connect ();
			break;
	}
	
	Print ();
}

HCI *BluntServer::SetupHCILayer (TObjectId port, UChar* addr, UByte psRepMode, UByte psMode)
{
	HCI* hci;
	int i;

	LOG ("BluntServer::SetupHCILayer\n");
	for (hci = NULL, i = 1; hci == NULL && i < fNumHandlers; i++) {
		if (memcmp (((HCI *) fHandlers[i])->fPeerAddress, addr, 6) == 0) hci = (HCI *) fHandlers[i];
	}
	if (hci == NULL) {
		hci = new HCI ();
		LOG ("  HCI: %08x\n", hci);
		hci->fConnectionHandle = -1;
		hci->fTool = port;
		hci->fConnectionState = CONN_CONNECT;
		memcpy (hci->fPeerAddress, addr, 6);
		hci->fControlChannel = (HCI *) fHandlers[0];
		hci->fPSRepMode = psRepMode;
		hci->fPSMode = psMode;
		AddHandler (hci);
	}
	
	return hci;
}

L2CAP* BluntServer::SetupL2CAPLayer (TObjectId port, Short protocol, HCI* hci)
{
	L2CAP* controlL2cap;
	L2CAP* l2cap;
	int i;

	LOG ("BluntServer::SetupL2CAPLayer\n");
	
	if (hci->fNumHandlers == 0) {
		controlL2cap = new L2CAP ();
		controlL2cap->fLocalCID = CID_SIGNAL;
		controlL2cap->fRemoteCID = CID_SIGNAL;
		hci->AddHandler ((Handler *) controlL2cap, this);
	} else {
		controlL2cap = (L2CAP *) hci->fHandlers[0];
	}
	controlL2cap->fTool = port;
	controlL2cap->fTargetProtocol = protocol;
	LOG (" L2CAP control: %08x\n", controlL2cap);
	
	for (l2cap = NULL, i = 1; l2cap == NULL && i < hci->fNumHandlers; i++) {
		if (((L2CAP *) hci->fHandlers[i])->fProtocol == protocol) l2cap = (L2CAP *) hci->fHandlers[i];
	}
	if (l2cap == NULL) {
		l2cap = new L2CAP ();
		l2cap->fLocalCID = controlL2cap->GetNewCID ();
		l2cap->fProtocol = protocol;
		l2cap->fConnectionState = CONN_CONNECT;
		hci->AddHandler ((Handler *) l2cap, this);
	}
	l2cap->fTool = port;
	LOG (" L2CAP data: %08x S: %d Prot: %d\n", l2cap, l2cap->fLocalCID, l2cap->fProtocol);
	
	return l2cap;
}

SDP* BluntServer::SetupSDPLayer (TObjectId port, L2CAP* l2cap)
{
	SDP* sdp;
	int i;
	
	LOG ("BluntServer::SetupSDPLayer\n");
	for (sdp = NULL, i = 0; sdp == NULL && i < l2cap->fNumHandlers; i++) {
		if (l2cap->fHandlers[i]->fHandlerType == H_SDP) sdp = (SDP *) l2cap->fHandlers[i];
	}
	if (sdp == NULL) {
		sdp = new SDP ();
		sdp->fConnectionState = CONN_CONNECT;
		l2cap->AddHandler ((Handler *) sdp, this);
	}
	sdp->fTool = port;
	return sdp;
}

RFCOMM* BluntServer::SetupRFCOMMLayer (TObjectId port, UByte comPort, L2CAP* l2cap)
{
	RFCOMM* rfcomm;
	RFCOMM* controlRfcomm;
	
	LOG ("BluntServer::SetupRFCOMMLayer\n");
	if (l2cap->fNumHandlers == 0) {
		controlRfcomm = new RFCOMM ();
		controlRfcomm->fTool = port;
		controlRfcomm->fDLCI = 0;
		controlRfcomm->fPort = 0;
		controlRfcomm->fTargetPort = comPort;
		l2cap->AddHandler ((Handler *) controlRfcomm, this);
	} else {
		controlRfcomm = (RFCOMM *) l2cap->fHandlers[0];
	}
	LOG (" RFCOMM control: %08x\n", controlRfcomm);
	rfcomm = new RFCOMM ();
	rfcomm->fPort = comPort;
	rfcomm->fDLCI = comPort << 1;
	rfcomm->fTool = port;
	rfcomm->fConnectionState = CONN_CONNECT;
	l2cap->AddHandler ((Handler *) rfcomm, this);
	LOG (" RFCOMM data: %08x Port: %d\n", rfcomm, rfcomm->fPort);

	return rfcomm;
}

void BluntServer::InitiatePairing (BluntInitiatePairingCommand* command)
{
	HCI* hci;

	LOG ("BluntServer::InitiatePairing\n");
	
	hci = SetupHCILayer (0, command->fBdAddr, command->fPSRepMode, command->fPSMode);
	hci->Pair (command->fBdAddr, command->fPIN);
}

void BluntServer::InitiateServiceRequest (BluntServiceRequestCommand* command)
{
	HCI* hci;
	L2CAP* l2cap;
	SDP* sdp;

	LOG ("BluntServer::InitiateServiceRequest\n");
	hci = SetupHCILayer (0, command->fBdAddr, command->fPSRepMode, command->fPSMode);
	l2cap = SetupL2CAPLayer (0, PSM_SDP, hci);
	sdp = SetupSDPLayer (0, l2cap);
	Print ();
	sdp->Connect ();
}

void BluntServer::Output (UByte* data, ULong size, Boolean packetStart)
{
	ULong n, processedBytes;
	int i;
	UByte c;
	
	LOG2 ("BluntServer::Output: %08x %d\n", data, size);

	i = 0;
	while (i < size && i < 96) {
		LOG2 ("%02x ", data[i]);
		i++;
		if ((i + 1) % 16 == 0) LOG2 ("\n");
	}
	LOG2 ("\n");
	
	if (packetStart) {
		fOutstandingPackets++;
	}
	
	n = size;
	fTxDMABuffer->CopyIn (data, &n, true, n);
	processedBytes = size - n;

	if (DMA) {
		fChip->SetTxDTransceiverEnable (false);
		fChip->TxDMAControl (kDMAStart);
		fChip->SetTxDTransceiverEnable (true);
	} else {
		fChip->SetTxDTransceiverEnable (true);
		i = 0;
		while (fChip->TxBufEmpty () && fTxDMABuffer->BufferCount () > 0) {
			fTxDMABuffer->GetNextByte (&c);
			i++;
			fChip->PutByte (c);
		}
	}
}

void BluntServer::SendEvent (BluntEvent* event)
{
	TUPort* port;
	
	LOG ("BluntServer::SendEvent %08x\n", event);
	port = GetNewtTaskPort ();
	port->Send (event, (void *) event, event->GetSizeOf ());
}

void BluntServer::SetTimer (Handler *handler, int milliSecondDelay, void *userData)
{
	BluntTimerEvent* timer = new BluntTimerEvent (kResultOk, handler, userData);
	TTime delay;
	
	LOG ("BluntServer::SetTimer %08x (%d) %d\n", handler, handler->fHandlerType, milliSecondDelay);
	delay = GetGlobalTime () + TTime (milliSecondDelay, kMilliseconds);
	fPort.Send (timer, (void *) timer, sizeof (*timer), kNoTimeout, &delay, M_TIMER);
}

TObjectId BluntServer::Port ()
{
	TUNameServer nameServer;
	ULong id;
	ULong spec;

	nameServer.Lookup ("BluntServer", "TUPort", &id, &spec);
	return id;
}

void BluntServer::AddHandler (Handler *handler)
{
	LOG ("BluntServer::AddHandler: %08x\n", handler);
	handler->fServer = this;
	fHandlers = (Handler**) realloc (fHandlers, (fNumHandlers + 1) *
		sizeof (Handler *));
	fHandlers[fNumHandlers] = handler;
	fNumHandlers++;
}

void BluntServer::RemoveHandler (Handler *handler)
{
	int i;
	
	LOG ("BluntServer::RemoveHandler %08x\n", handler);
	for (i = 0; i < fNumHandlers; i++) {
		if (fHandlers[i] == handler) {
			memmove (&fHandlers[i], &fHandlers[i + 1],
				sizeof (Handler *) + fNumHandlers - 1);
			fNumHandlers--;
			if (handler->fHandlerType == H_HCI) {
				delete (HCI *) handler;
			} else {
				delete handler;
			}
			break;
		}
	}
}

void BluntServer::DeferredRemoveHandler (Handler *handler)
{
	int i;
	
	LOG ("BluntServer::DeferredRemoveHandler %08x\n", handler);
	for (i = 0; i < fNumHandlers; i++) {
		if (fHandlers[i] == handler) {
			fHandlers[i]->fConnectionState = CONN_REMOVE;
			break;
		}
	}
}

void BluntServer::Print ()
{
	int i;
	
	LOGX ("BluntServer::Print\n");
	for (i = 0; i < fNumHandlers; i++) fHandlers[i]->Print (4);
}

/*
void BluntServer::RxCAvailIntHandler (void)
{
	ULong s;
	
	while (fChip->RxBufFull () && fInputCurrent < MAX_INPUT) {
		s = fChip->GetRxErrorStatus ();
		if (s != 0) fRxStatus = s;
		fInputBuffer[fInputCurrent++] = fChip->GetByte ();
	}
	SendForInterrupt (fPort, fIntMessage.GetMsgId (), 0, NULL, 0, M_DATA);
}
*/

/*
void BluntServer::HandleData ()
{
	int i;
	ULong lastAmount;
	
	LOG2 ("BluntServer::HandleData %08x %d\n", fInputBuffer, fInputCurrent);
	lastAmount = 0;
	do {
		lastAmount = fInputCurrent;
		for (i = 0; fInputCurrent > 0 && i < fNumHandlers; i++) {
			LOG2 ("BluntServer Handler: %d (%08x), data: %d bytes\n", i, fHandlers[i], fInputCurrent);
			((HCI *) fHandlers[i])->HandleData (fInputBuffer, fInputCurrent);
		}
	} while (lastAmount != fInputCurrent);

	if (fInputCurrent > 0) {
		do {
			lastAmount = fInputCurrent;
			((HCI *) fHandlers[0])->RemoveUnhandledPackets (fInputBuffer, fInputCurrent);
		} while (fInputCurrent > 0 && lastAmount != fInputCurrent);
	}
}
void BluntServer::InitiatePairing (BluntInitiatePairingCommand* command)
{
	HCI* hci;

	LOG ("BluntServer::InitiatePairing\n");
	
	// set up the channel we're connecting over
	hci = new HCI ();
	hci->fConnectionHandle = -1;
	hci->fConnectionState = CONN_CONNECT;
	memcpy (hci->fPeerAddress, command->fBdAddr, 6);
	hci->fControlChannel = (HCI *) fHandlers[0];
	AddHandler (hci);
}
*/


/*
x:init(); x:start();
e := {_proto: @383}; e:Instantiate (e, {label: "rfcm", type: 'service, opCode: 512}); e:bind(nil, nil)
e:connect(nil, {async: true, reqtimeout: 20000, completionscript: func (a, b, c) begin end});
e:disconnect(nil, nil);
e:unbind(nil);
x:stop();
*/
