#include <AEvents.h>
#include <AEventHandler.h>
#include <BufferSegment.h>
#include <CommManagerInterface.h>
#include <CommTool.h>
#include <SerialOptions.h>
#include <CommOptions.h>

#include "BluntServer.h"
#include "HCI.h"
#include "SDP.h"
#include "L2CAP.h"
#include "RFCOMM.h"

#define PF					0x10
#define COMMAND			 	0x02
#define RESPONSE			0x00
#define EA					0x01

RFCOMM::RFCOMM (void)
{
	fLogLevel = 1;
	LOG ("RFCOMM::RFCOMM\n");
	
	fHandlerType = H_RFCOMM;
	fInitiator = false;
	fOutputPacket = NULL;
	fInputPacket = NULL;
	fState = RFCOMM_IDLE;
	fConfigured = false;
	fRemoteMTU = 0;
	fDLCI = 0;
	CreateCRCTable ();
}
	
RFCOMM::~RFCOMM (void)
{
}
	
#pragma mark -

Boolean RFCOMM::HandleData (UByte *data, ULong length)
{
	Boolean r;
	UByte control;
	Byte DLCI;
	
	DLCI = data[0] >> 2;
	LOG ("RFCOMM::HandleData %08x (%d %d)\n", data, DLCI, fDLCI);
	r = false;
	if (DLCI == fDLCI) {
		if ((data[2] & 0x01) != 0x01) {
			fInputPacketLength = (data[2] >> 1) + (data[3] << 7);
			fInputPacket = data + 4;
		} else {
			fInputPacketLength = data[2] >> 1;
			fInputPacket = data + 3;
		}
		
		if ((data[0] & COMMAND) == COMMAND) fCR = true; else fCR = false;
		if ((data[1] & PF) == PF) fPF = true; else fPF = false;
		
		control = data[1] & 0xef;
		
		LOG (" Address: %d Control: %02x Length: %d\n",
			DLCI, control, fInputPacketLength);
			
		switch (control) {
			case CTRL_SABM:
				if (fPF) SetAsynchronousBalancedMode ();
				break;
			case CTRL_UA:
				UnnumberedAcknowlegdement ();
				break;
			case CTRL_DM:
				DisconnectedMode ();
				break;
			case CTRL_DISC:
				if (fPF) Disconnect ();
				break;
			case CTRL_UIH:
				r = UnnumberedInformation ();
				break;
		}
		r = true;
	}
	
	return r;
}

void RFCOMM::HandleTimer (void *userData)
{
	LOG ("RFCOMM::HandleTimer\n");
}

void RFCOMM::Connect (void)
{
}

void RFCOMM::CompleteConnection (void)
{
	RFCOMM *r, *rfcomm;
	BluntConnectionCompleteEvent *e;
	int i;
	
	LOG ("RFCOMM::CompleteConnection (%08x)\n", this);
	if (fDLCI == 0) {
		for (rfcomm = NULL, i = 1; rfcomm == NULL && i < fParentHandler->fNumHandlers; i++) {
			r = (RFCOMM *) fParentHandler->fHandlers[i];
			LOG ("  %d %d %d\n", i, r->fPort, r->fConnectionState);
			if (r->fPort == fTargetPort && r->fConnectionState == CONN_CONNECT) {
				rfcomm = r;
			}
		}
		if (rfcomm != NULL) {
			LOG ("  Connecting %08x (%d)\n", r, r->fPort);
			r->SndSetAsynchronousBalancedMode ();
		}
	} else {
		TUPort p (fTool);
		fConnectionState = CONN_CONNECTED;
		fState = RFCOMM_IDLE;
		
		e = new BluntConnectionCompleteEvent (noErr);
		e->fHCIHandle = ((HCI *) fParentHandler->fParentHandler)->fConnectionHandle;
		e->fL2CAPLocalCID = ((L2CAP *) fParentHandler)->fLocalCID;
		e->fL2CAPRemoteCID = ((L2CAP *) fParentHandler)->fRemoteCID;
		e->fRFCOMMPort = fPort;
		p.Send ((TUAsyncMessage *) e, e, sizeof (BluntConnectionCompleteEvent), kNoTimeout, nil, BLUNT_MSG_TYPE);
	}
}

void RFCOMM::SetAsynchronousBalancedMode ()
{
	LOG ("RFCOMM::SetAsynchronousBalancedMode\n");

	fNegotiationComplete = false;
	fInitiator = false;
	SndUnnumberedAcknowlegdement ();
}

void RFCOMM::UnnumberedAcknowlegdement ()
{
	Short mtu;
	
	LOG ("RFCOMM::UnnumberedAcknowlegdement (%d)\n", fDLCI);
	if (fRemoteMTU > 0) {
		mtu = MIN(RFCOMM_MTU_LEN, fRemoteMTU - 8);
	} else {
		mtu = RFCOMM_MTU_LEN;
	}
	if (fDLCI == 0) {
		SndMPXParameterNegotiation (fTargetPort << 1, 0, mtu, 0);
	} else {
		SndMPXModemStatusCommand (fDLCI, true, false, 0x8c);
		CompleteConnection ();
	}
}

void RFCOMM::DisconnectedMode ()
{
	LOG ("RFCOMM::DisconnectedMode\n");
}

void RFCOMM::Disconnect ()
{
	LOG ("RFCOMM::Disconnect\n");

	SndUnnumberedAcknowlegdement ();
	fState = RFCOMM_IDLE;
}

int RFCOMM::UnnumberedInformation ()
{
	int r;
	
	LOG ("RFCOMM::UnnumberedInformation\n");
	r = noErr;
	if (fDLCI == 0) {
		ProcessMultiplexerCommands ();
	} else {
		r = ProcessRFCOMMData ();
	}
	return r;
}

void RFCOMM::SndSetAsynchronousBalancedMode ()
{
	UByte response[4];

	LOG ("RFCOMM::SndSetAsynchronousBalancedMode %08x (%d)\n", this, fDLCI);
	fState = RFCOMM_CONNECT;

	fInitiator = true;
	response[0] = (fDLCI << 2) | COMMAND | EA;
	response[1] = CTRL_SABM | PF;
	response[2] = (0 << 1) | EA;
	response[3] = CalculateCRC (response, 3);
	SndData (response, sizeof (response));
}

void RFCOMM::SndUnnumberedAcknowlegdement ()
{
	UByte response[4];

	LOG ("RFCOMM::SndUnnumberedAcknowlegdement\n");
	
	if (fDLCI != 0) {
		fState = RFCOMM_CONNECTED;
	}
	response[0] = (fDLCI << 2) | COMMAND | EA;
	response[1] = CTRL_UA | PF;
	response[2] = (0 << 1) | EA;
	response[3] = CalculateCRC (response, 3);
	SndData (response, sizeof (response));
}

void RFCOMM::SndDisconnectedMode ()
{
	LOG ("RFCOMM::SndDisconnectedMode\n");
}

void RFCOMM::SndDisconnect ()
{
	UByte response[4];

	LOG ("RFCOMM::SndDisconnect\n");

	if (fState == RFCOMM_CONNECTED) {
		response[0] = (fDLCI << 2) | COMMAND | EA;
		response[1] = CTRL_DISC | PF;
		response[2] = (0 << 1) | EA;
		response[3] = CalculateCRC (response, 3);
		SndData (response, sizeof (response));
	}
}

void RFCOMM::SndData (UByte *data, Short length)
{
	LOG ("RFCOMM::SndData\n");

	((L2CAP *) fParentHandler)->SndPacketHeader (length);
	fServer->Output (data, length);
}

void RFCOMM::SndUnnumberedInformation (UByte *data, Short len)
{
	UByte header[4];
	UByte crc;
	Short headerLen;
	
	LOG ("RFCOMM::SndUnnumberedInformation (%d %d)\n", fDLCI, len);

	header[0] = (fDLCI << 2) | COMMAND | EA; 
	header[1] = CTRL_UIH;
	if (len > 127) {
		header[2] = (len & 0x007f) << 1;
		header[3] = (len & 0x7f80) >> 7;
		headerLen = 4;
		fServer->Output (header, 4);
	} else {
		header[2] = (len << 1) | EA;
		headerLen = 3;
	}
	crc = CalculateCRC (header, 2);

	((L2CAP *) fParentHandler)->SndPacketHeader (headerLen + len + sizeof (crc));
	fServer->Output (header, headerLen);
	fServer->Output (data, len);
	fServer->Output (&crc, sizeof (crc));
}

void RFCOMM::ProcessMultiplexerCommands (void)
{
	Byte command;
	Short length;
	Boolean cr;
	
	LOG ("RFCOMM::ProcessMultiplexerCommands\n");
	
	if ((fInputPacket[0] & 0x01) == 0x01) {
		cr = (fInputPacket[0] & COMMAND) == COMMAND ? true : false;
		command = fInputPacket[0] >> 2;
		
		switch (command) {
			case MPX_PN:
				MPXParameterNegotiation (cr);
				break;
			case MPX_RPN:
				MPXRemotePortNegotiation (cr);
				break;
			case MPX_MSC:
				MPXModemStatusCommand (cr);
				break;
			case MPX_PSC:
			case MPX_CLD:
			case MPX_Test:
			case MPX_FCon:
			case MPX_FCOff:
			case MPX_NSC:
			case MPX_RLS:
			case MPX_SNC:
				LOG ("  %02x\n", command);
				break;
		}
	}
}

int RFCOMM::ProcessRFCOMMData (void)
{
	Short len;
	int r;

	LOG ("RFCOMM::ProcessRFCOMMData (%d)\n", fInputPacketLength);
	r = noErr;
	if (fInputPacketLength > 0) {
		LOG ("  %02x (%c)\n", fInputPacket[0], fInputPacket[0]);
//		r = fL2CAP->fServer->HandleInputData (fInputPacket, fInputPacketLength); XXX
	}
	
	return r;
}

void RFCOMM::MPXParameterNegotiation (Boolean command)
{
	Byte negotiationDLCI;
	Byte flowControl;
	UShort frameSize;
	Byte windowSize;
	
	LOG ("RFCOMM::MPXParameterNegotiation\n");
	
	negotiationDLCI = fInputPacket[2];
	flowControl = fInputPacket[3] >> 4;
	frameSize = fInputPacket[6] + (fInputPacket[7] << 8);
	windowSize = fInputPacket[9];
	LOG ("  DLCI: %d Flow: %02x Frame size: %d, Window: %d\n",
		negotiationDLCI, flowControl, frameSize, windowSize);

	if (command) {
		SndMPXParameterNegotiation (negotiationDLCI, 0, 667, 0);
		fNegotiationComplete = true;
	} else {
		CompleteConnection ();
	}
}

void RFCOMM::MPXRemotePortNegotiation (Boolean command)
{
	LOG ("RFCOMM::MPXRemotePortNegotiation\n");
}

void RFCOMM::MPXModemStatusCommand (Boolean command)
{
	Byte DLCI;

	DLCI = fInputPacket[2] >> 2;
	LOG ("RFCOMM::MPXModemStatusCommand (%d %d %d 0x%02x)\n",
		DLCI, command, fPF, fInputPacket[3]);
	if (command) {
		SndMPXModemStatusCommand (DLCI, false, fPF, fInputPacket[3]);
	}
}

void RFCOMM::SndMPXParameterNegotiation (Byte negotiationDLCI, Byte flowControl, Short frameSize, Byte windowSize)
{
	UByte response[14];

	LOG ("RFCOMM::SndMPXParameterNegotiation (%d %d %d)\n", fInitiator, negotiationDLCI, frameSize);

	response[0] = (0 << 2) | (fInitiator ? COMMAND : RESPONSE) | EA; 
	response[1] = CTRL_UIH | PF; 
//	response[1] = CTRL_UIH | 0; // T68i XXX
	response[2] = ((sizeof (response) - 4) << 1) | EA;
	
	response[3] = (MPX_PN << 2) | (fInitiator ? COMMAND : RESPONSE) | EA;
	response[4] = ((sizeof (response) - 6) << 1) | EA;
	response[5] = negotiationDLCI;
	response[6] = flowControl << 4 | 0; // CL flow control & I-Frames
	response[7] = 0; // Priority
	response[8] = 0; // ACK timer
	response[9] = frameSize & 0x00ff; // Frame size
	response[10] = frameSize >> 8;
	response[11] = 0; // Retransmissions;
	response[12] = windowSize; // Window size
	
	response[13] = CalculateCRC (response, 2);
	SndData (response, sizeof (response));
}

void RFCOMM::SndMPXRemotePortNegotiation (void)
{
	LOG ("RFCOMM::SndMPXRemotePortNegotiation\n");
}

void RFCOMM::SndMPXModemStatusCommand (Byte DLCI, Boolean isCommand, Boolean pf, Byte status)
{
	UByte response[8];

	LOG ("RFCOMM::SndMPXModemStatusCommand (%d %d %d)\n", DLCI, fInitiator, isCommand);

	response[0] = (0 << 2) | (fInitiator ? COMMAND : RESPONSE) | EA; 
	response[1] = CTRL_UIH | (pf ? PF : 0);
	response[2] = ((sizeof (response) - 4) << 1) | EA;
	
	response[3] = (MPX_MSC << 2) | (isCommand ? COMMAND : RESPONSE) | EA;
	response[4] = ((sizeof (response) - 6) << 1) | EA;
	response[5] = (DLCI << 2) | 0x02 | EA;
	response[6] = status | 0x01;
	
	response[7] = CalculateCRC (response, 2);
	SndData (response, sizeof (response));
}

void RFCOMM::CreateCRCTable ()
{
	int i ,j;
	UByte data;
	UByte code_word = 0xe0; // pol = x8+x2+x1+1
	UByte sr = 0;			// Shiftregister initiated to zero
	
	for (j = 0; j < 256; j++) {
		data = (UByte) j;
		for (i = 0; i < 8; i++) {
			if ((data & 0x1) ^ (sr & 0x1)) {
				sr >>= 1;
				sr ^= code_word;
			} else {
				sr >>= 1;
			}
			data >>= 1;
			sr &= 0xff;
		}
		fCRCTable[j] = sr;
		sr = 0;
	} 
}

UByte RFCOMM::CalculateCRC (UByte *data, Byte length)
{
	UByte fcs = 0xff;

	while (length--) {
		fcs = fCRCTable[fcs ^ *data++];
	}
	
	return 0xff-fcs;
} 
